# -*- coding: utf-8 -*-
import contextlib
import os
from pathlib import Path
from string import Template
from typing import Union, List, MutableMapping

__all__ = (
    'load_env',
    'load_dotenv',      # alias
    'unquote',
)

DEFAULT_ENVKEY = 'DOTENV'
DEFAULT_DOTENV = '.env'


def unquote(line, quotes='"\''):
    if line and line[0] in quotes and line[-1] == line[0]:
        line = line[1:-1]
    return line


def _env_default(environ: MutableMapping[str, str], key: str, val: str, overwrite: bool=False):
    if key and val:
        if overwrite or key not in environ:
            environ[key] = val


def _env_export(environ: MutableMapping[str, str], key: str, val: str, overwrite: bool=False):
    if key and val:
        if overwrite or key not in environ:
            environ[key] = val
            os.environ[key] = val


def _env_files(env_file: str, search_path: List[Path], parents: bool, errors: bool) -> List[str]:
    """ expand env_file with full search path, optionally parents as well """

    searched = []
    for path in search_path:
        path = path.resolve()
        if not path.is_dir():
            path = path.parent
        searched.append(path)
        paths = [path] + list(path.parents)
        # search a path and parents
        for sub_path in paths:
            # stop at first found, or ...
            # fail fast unless searching parents
            env_path = os.path.join(sub_path, env_file)
            if os.access(env_path, os.R_OK):
                yield env_path
            elif not parents:
                break
    if errors:
        raise FileNotFoundError(f"{env_file} in {[s.as_posix() for s in searched]}")
    else:
        yield env_file


@contextlib.contextmanager
def open_env(path: Union[str, Path]):
    """same as open, allow monkeypatch"""
    fp = None
    try:
        fp = open(path, 'r')
        yield fp
    finally:
        if fp:
            fp.close()


ENV_COMMANDS = {
    'export': _env_export,
}


def _process_env(env_file: str, search_path: List[Path], environ: MutableMapping[str, str], overwrite: bool,
                 parents: bool, errors: bool) -> MutableMapping[str, str]:
    """ search for any env_files in given dir list and populate environ dict
    :param env_file: base environment file name to use
    :param search_path: list of one or more paths to search
    :param environ:  environment to update
    :param overwrite:
    :param parents:
    """

    def process_line(env_path: Path, lineno: int, string: str):
        """ process a single line """
        func, key, val = _env_default, None, None
        parts = string.split('=', 1)
        if len(parts) == 2:
            key, val = parts
        elif len(parts) == 1:
            key = parts[0]
        if key:
            words = key.split(maxsplit=1)
            if len(words) > 1:
                command, key = words
                try:
                    func = ENV_COMMANDS[command]
                except KeyError:
                    if errors:
                        print(f"unknown command {command} {env_path.as_posix()}({lineno})", file=sys.stderr)
        return func, unquote(key), unquote(val)

    for env_path in _env_files(env_file, search_path, parents, errors):
        # insert PWD as container of env file
        env_path = Path(env_path).resolve()
        environ['PWD'] = str(env_path.parent)
        try:
            with open_env(env_path) as f:
                lineno = 0
                for line in f.readlines():
                    line = line.strip()
                    lineno += 1
                    if line and line[0] != '#':
                        func, key, val = process_line(env_path, lineno, line)
                        if func is not None:
                            func(environ, key, val, overwrite=overwrite)
        except FileNotFoundError:
            if errors:
                raise
    return environ


def _post_process(environ: MutableMapping[str, str]) -> MutableMapping[str, str]:
    """ post-process the variables using ${substitutions} """
    for env_key, env_val in environ.items():
        if all(v in env_val for v in ('${', '}')):  # looks like template
            # ignore anything that does not resolve, don't throw an exception!
            val = Template(env_val).safe_substitute(environ)
            if val != env_val:  # don't update unless we need to
                environ[env_key] = val
    return environ


def _update_os_env(environ: MutableMapping[str, str]) -> MutableMapping[str, str]:
    """ back-populate changed variables to the environment """
    for env_key, env_val in environ.items():
        if env_val != os.environ.get(env_key):
            os.environ[env_key] = env_val
    return os.environ


def load_env(env_file: str = None, search_path: Union[None, Union[List[str], List[Path]], str] = None,
             overwrite: bool = False, parents: bool = False, update: bool = True, errors: bool = False,
             environ: MutableMapping[str, str] = None) -> MutableMapping[str, str]:
    """
    Loads one or more .env files with optional nesting, updating os.environ
    :param env_file: name of the environment file (.env or $ENV default)
    :param search_path: single or list of directories in order of precedence - str, bytes or Path
    :param overwrite: whether to overwrite existing values
    :param parents: whether to search upwards until a file is found
    :param update: option to update os.environ, default=True
    :param environ: environment mapping to process
    :returns the new environment
    """
    if environ is None:
        environ = os.environ
    if not env_file:
        env_file = environ.get(DEFAULT_ENVKEY, DEFAULT_DOTENV)

    # insert this as a useful default
    environ['CWD'] = str(Path.cwd().resolve(strict=True))

    # determine where to search
    if search_path is None:
        import inspect
        frame = inspect.stack()[1]
        search_path = ['.', frame.filename]
    elif isinstance(search_path, (str, bytes, Path)):
        search_path = [search_path]
    # convert to array of Path for use internally
    search_path = [Path(p) for p in search_path]
    # if overwriting, traverse path in reverse order so first .env files have priority
    if overwrite:
        search_path.reverse()

    # slurp up the environment files found and
    # post process values for template variables
    environ = _post_process(_process_env(env_file, search_path, environ.copy(), overwrite, parents, errors))
    # optionally update the actual environment
    return _update_os_env(environ) if update else environ


load_dotenv = load_env



# -*- coding: utf-8 -*-
import contextlib
import io
import envex
import pytest


TEST_ENV = [
    '# This is an example .env file',
    'DATABASE_URL=postgresql://username:password@localhost/database_name',
    'CACHE_URL=memcache://localhost:11211',
    'REDIS_URL=redis://localhost:6379/5',
    'QUOTED_VALUE="some double quoted value"',
    'INTVALUE=225',
    'FLOATVALUE=54.92',
    'BOOLVALUETRUE=True',
    'BOOLVALUEFALSE=off',
    'LISTOFQUOTEDVALUES=1,"two",3,\'four\'',
    'ALISTOFIPS=::1,127.0.0.1,mydomain.com',
]


@contextlib.contextmanager
def dotenv(ignored):
    _ = ignored
    yield io.StringIO("\n".join(TEST_ENV))


def test_env_wrapper():
    env = envex.Env()
    assert 'HOME' in env
    assert 'USER' in env


def test_env_call():
    env = envex.Env()
    var, val = 'MY_VARIABLE', 'MY_VARIABLE_VALUE'
    assert var not in env
    value = env(var)
    assert value is None
    value = env(var, val)
    assert value == val
    assert var in env
    assert env[var] == val


def test_env_int(monkeypatch):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.Env(readenv=True)
    assert env.int('INTVALUE', default=99) == 225
    assert env('INTVALUE', default=99, type=int) == 225
    assert env.int('DEFAULTINTVALUE', default=981) == 981
    assert env('DEFAULTINTVALUE', default=981, type=int) == 981
    assert env('DEFAULTINTVALUE', type=int) == 981


def test_env_float(monkeypatch):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.Env(readenv=True)
    assert env.float('FLOATVALUE', default=99.9999) == 54.92
    assert env('FLOATVALUE', default=99.9999, type=float) == 54.92
    assert env.float('DEFAULTFLOATVALUE', default=83.6) == 83.6
    assert env('DEFAULTFLOATVALUE', default=83.6, type=float) == 83.6
    assert env('DEFAULTFLOATVALUE', type=float) == 83.6


def test_is_true():
    env = envex.Env()
    assert env.is_true(1)
    assert env.is_true('1')
    assert not env.is_true(0)
    assert not env.is_true('0')
    assert not env.is_true(b'0')
    assert not env.is_true(False)
    assert not env.is_true('False')
    assert not env.is_true(None)


def test_env_bool(monkeypatch):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.Env(readenv=True)
    assert env.bool('BOOLVALUETRUE', default=False)
    assert env.bool('DEFAULTBOOLVALUETRUE', default=True)
    assert env('DEFAULTBOOLVALUETRUE', default=True, type=bool)
    assert not env.bool('BOOLVALUEFALSE', default=True)
    assert not env.bool('DEFAULTBOOLVALUEFALSE', default=False)
    assert not env('DEFAULTBOOLVALUEFALSE', type=bool)


def test_env_list(monkeypatch):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.Env(readenv=True)

    result = env.list('ALISTOFIPS')
    assert isinstance(result, list)
    assert len(result) == 3
    assert result == ['::1', '127.0.0.1', 'mydomain.com']

    result = env('ALISTOFIPS', type=list)
    assert isinstance(result, list)
    assert len(result) == 3
    assert result == ['::1', '127.0.0.1', 'mydomain.com']

    result = env.list('LISTOFQUOTEDVALUES')
    assert isinstance(result, list)
    assert len(result) == 4
    assert result == ['1', 'two', '3', 'four']

    result = env('LISTOFQUOTEDVALUES', type=list)
    assert isinstance(result, list)
    assert len(result) == 4
    assert result == ['1', 'two', '3', 'four']


def test_env_iter(monkeypatch):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.Env(readenv=True, update=False)

    # test items() itself (returned by __iter__)
    for var, val in env.items():
        assert isinstance(var, str)
        assert isinstance(val, str)

    # test __iter__ via list()
    for var, val in list(env):
        assert isinstance(var, str)
        assert isinstance(val, str)

    # test __iter__ via dict()
    for var, val in dict(env).items():
        assert isinstance(var, str)
        assert isinstance(val, str)


def test_env_exception():
    class MyException(Exception):
        pass
    env = envex.Env(exception=MyException)
    with pytest.raises(MyException):
        _ = env['UNDEFINEDVARIABLE']


def test_env_export():
    env = envex.Env()
    assert 'MYVARIABLE' not in env
    env.export(MYVARIABLE='somevalue')
    assert env['MYVARIABLE'] == 'somevalue'
    env.export(MYVARIABLE=None)
    with pytest.raises(KeyError):
        _ = env['MYVARIABLE']

    values = dict(MYVARIABLE='somevalue', MYVARIABLE2=1, MYVARIABLE3='...')

    env.export(values)
    for k, v in values.items():
        assert env[k] == str(v)
    env.export({k: None for k in values.keys()})
    assert not env.is_any_set([k for k in values.keys()])

    env.export(**values)
    for k, v in values.items():
        assert env[k] == str(v)
    env.export({k: None for k in values.keys()})
    assert not env.is_any_set([k for k in values.keys()])

    import os

    env.export(**values)
    for k, v in values.items():
        assert os.environ[k] == str(v)


def test_env_contains(monkeypatch):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.Env()
    # must be explicitly read in
    env.read_env()

    assert 'DATABASE_URL' in env
    assert env['DATABASE_URL'] == "postgresql://username:password@localhost/database_name"
    assert 'CACHE_URL' in env
    assert env['CACHE_URL'] == "memcache://localhost:11211"
    assert 'REDIS_URL' in env
    assert env['REDIS_URL'] == "redis://localhost:6379/5"



# -*- coding: utf-8 -*-
"""
Wrapper around os.environ
"""
import re
from typing import Any, MutableMapping
from urllib.parse import urlparse, urlunparse, parse_qs, unquote_plus
from .dot_env import load_env, unquote


class Env:
    """
    Wrapper around os.environ with .env enhancement` and django support
    """
    _BOOLEAN_TRUE_STRINGS = ('T', 't', '1', 'on', 'ok', 'Y', 'y', 'en')
    _BOOLEAN_TRUE_BYTES = (s.encode('utf-8') for s in _BOOLEAN_TRUE_STRINGS)
    _EXCEPTION_CLS = KeyError

    @staticmethod
    def os_env():
        import os
        return os.environ

    def __init__(self, *args, environ: MutableMapping[str, str] = None, exception=None, readenv=False, **kwargs):
        self._typemap = {
            str: self.get,
            int: self.int,
            bool: self.bool,
            float: self.float,
            list: self.list,
        }
        self._env = environ or self.os_env()
        self._env.update(args)
        if readenv:
            self.read_env(**kwargs)
        else:
            self._env.update(kwargs)
        self.exception = exception or self._EXCEPTION_CLS

    def read_env(self, **kwargs):
        """
        :param kwargs: see load_env
            env_file: str
            search_path: Union[None, Union[List[str], List[Path]], str]
            overwrite: bool
            parents: bool
            update: bool
        :   MutableMapping[str, str]
        """
        kwargs.setdefault('environ', self._env)
        self._env = load_env(**kwargs)

    @property
    def exception(self):
        return self._exception

    @exception.setter
    def exception(self, exc):
        if not issubclass(exc, Exception):
            raise ValueError(f'arg {exc} is not an exception class')
        self._exception = exc

    @property
    def env(self):
        return self._env

    def get(self, var, default=None):
        return self.env.get(var, default)

    def __call__(self, var, default=None, type=str):
        if default is not None and not self.is_set(var):
            self.set(var, default)
        try:
            return self._typemap[type](var, default=default)
        except KeyError:
            pass
        return self.get(var, default)

    def pop(self, var, default=None):
        val = self.get(var, default)
        self.unset(var)
        return val

    def set(self, var, value=None):
        self.env[var] = str(value) if value is not None else value

    def setdefault(self, var, value):
        return self.env.setdefault(var, str(value) if value is not None else value)

    def unset(self, var):
        if var in self.env:
            del self._env[var]

    def is_set(self, var):
        return var in self

    def is_all_set(self, *_vars):
        return all(v in self for v in _vars)

    def is_any_set(self, *_vars):
        return any(v in self for v in _vars)

    def int(self, var, default=None) -> int:
        val = self.get(var, default)
        return self._int(val)

    def float(self, var, default=None) -> float:
        val = self.get(var, default)
        return self._float(val)

    def bool(self, var, default=None) -> bool:
        val = self.get(var, default)
        return val if isinstance(val, (bool, int)) else self.is_true(val)

    def list(self, var, default=None) -> list:
        val = self.get(var, default)
        return val if isinstance(val, (list, tuple)) else self._list(val)

    def export(self, *args, **kwargs):
        import os

        for arg in args:
            if not isinstance(arg, (dict,)):
                raise TypeError('export() requires either dictionaries or keyword=value pairs')
            kwargs |= {k: v for k, v in arg.items()}
        if not args and not kwargs:
            kwargs = self._env
        for k, v in kwargs.items():
            k = str(k)
            try:
                if v is None:
                    self.unset(k)
                    del os.environ[k]
                else:
                    self.set(k, v)
                    os.environ[k] = str(v)
            except KeyError:
                ...
    @classmethod
    def _true_values(cls, val):
        return cls._BOOLEAN_TRUE_STRINGS if isinstance(val, str) else cls._BOOLEAN_TRUE_BYTES

    @classmethod
    def is_true(cls, val):
        if val in (None, False, '', 0, '0'):
            return False
        if not isinstance(val, (str, bytes)):
            return bool(val)
        true_vals = cls._true_values(val)
        return True if val and any([val.startswith(v) for v in true_vals]) else False

    @classmethod
    def _int(cls, val):
        return val if isinstance(val, int) else int(val) if val and str.isdigit(val) else 0

    @classmethod
    def _float(cls, val):
        return val if isinstance(val, float) else float(val) if val else 0

    @classmethod
    def _list(cls, val):
        return [] if val is None else [unquote(part) for part in re.split(r'\s*,\s*', str(val))]

    def __contains__(self, var):
        return str(var) in self.env

    def __setitem__(self, var: str, value: Any):
        self.set(var, value)

    def __getitem__(self, var):
        if var not in self:
            raise self.exception(f"Key '{var}' not found")
        return self.get(var)

    def __delitem__(self, var):
        self.unset(var)

    def items(self):
        for var, val in self.env.items():
            yield var, val

    def __iter__(self):
        return self.items()

    def check_var(self, var, default=None, raise_error=True):
        if not var:
            url = None
        else:
            url = self.get(var, default=default) if var else default
            if not url and raise_error:
                raise self.exception(f'Expected {var} is not set in environment')
        return '' if url is None else url

env = Env()



# -*- coding: utf-8 -*-
import contextlib

import envex
import io
import pytest


@pytest.fixture
def envmap():
    return {
        'FIRST': 'first-value',
        'SECOND': 'second-value',
        'THIRD': 'third-value',
        'FORTH': 'forth-value',
    }


@contextlib.contextmanager
def dotenv(ignored):
    _ = ignored
    yield io.StringIO("""
# This is an example .env file
SECOND=a-second-value
THIRD=altnernative-third
export FIFTH=fifth-value
COMBINED=${FIRST}:${THIRD}:${FIFTH}
DOUBLE_QUOTED="a quoted value"
SINGLE_QUOTED='a quoted value'
""")


def test_load_env(monkeypatch, envmap):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.load_env(search_path=__file__, environ=envmap)
    for var in envmap.keys():
        assert var in env
    assert 'FIFTH' in env
    assert env['COMBINED'] == 'first-value:third-value:fifth-value'


def test_load_env_overwrite(monkeypatch, envmap):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.load_env(search_path=__file__, environ=envmap, overwrite=True)
    for var in envmap.keys():
        assert var in env
    assert 'FIFTH' in env
    assert env['COMBINED'] == 'first-value:altnernative-third:fifth-value'


def test_quoted_value(monkeypatch, envmap):
    monkeypatch.setattr(envex.dot_env, 'open_env', dotenv)
    env = envex.load_env(search_path=__file__, environ=envmap)
    assert env['DOUBLE_QUOTED'] == 'a quoted value'
    assert env['SINGLE_QUOTED'] == 'a quoted value'
